@@ -847,7 +847,7 @@ fail:
  * waiting for it to complete.
  */
 static gpg_error_t iterate_crypto_once(struct client_s *client,
-	struct client_crypto_s *crypto, status_msg_t which)
+	struct crypto_s *crypto, status_msg_t which)
 {
     gpg_error_t rc = 0;
     goffset len = CRYPTO_BLOCKSIZE(crypto);
@@ -895,7 +895,7 @@ done:
 
 /* The crypto struct must be setup for iterations and .key. */
 gpg_error_t do_xml_encrypt(struct client_s *client,
-	struct client_crypto_s *crypto, const gchar *filename)
+	struct crypto_s *crypto, const gchar *filename)
 {
     goffset len = crypto->insize;
     gpointer inbuf;
@@ -1159,7 +1159,7 @@ do_write_file:
 }
 
 gpg_error_t update_save_flags(const gchar *filename,
-	struct client_crypto_s *crypto)
+	struct crypto_s *crypto)
 {
     gpg_error_t rc;
     guint64 iter;
@@ -3884,7 +3884,7 @@ gpg_error_t register_commands(assuan_context_t ctx)
 }
 
 gpg_error_t try_xml_decrypt(assuan_context_t ctx,
-	struct client_crypto_s *crypto, gpointer *dst, goffset *dst_len)
+	struct crypto_s *crypto, gpointer *dst, goffset *dst_len)
 {
     goffset insize, len;
     struct client_s *client = ctx ? assuan_get_pointer(ctx) : NULL;
