@@ -20,15 +20,18 @@
 #include <config.h>
 #endif
 
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include <gpg-error.h>
 #include <errno.h>
 #include <glib.h>
 #include "pwmd_error.h"
-#include "rcfile.h"
 #include "lock.h"
 #include "common.h"
+#include "rcfile.h"
 
-gboolean do_cache_push(const gchar *filename, const gchar *password);
+gboolean do_cache_push(const gchar *filename, const void *password, gsize);
 
 void clear_rcfile_keys()
 {
@@ -48,49 +51,60 @@ void clear_rcfile_keys()
     g_strfreev(groups);
 }
 
-static gboolean _getline(const gchar *file, gchar **result, gpg_error_t *rc)
+static gpg_error_t _getline(struct crypto_s *crypto, const gchar *file)
 {
-    FILE *fp;
-    gchar buf[LINE_MAX] = {0}, *p;
-    gchar *str = NULL;
-    gint len;
+    gint fd;
+    struct stat st;
+    gpg_error_t rc = 0;
 
-    *rc = 0;
+    if (stat(file, &st) == -1)
+	return gpg_error_from_errno(errno);
 
-    if ((fp = fopen(file, "r")) == NULL) {
-	*rc = gpg_error_from_syserror();
-	return FALSE;
+    crypto->tkey_len = st.st_size;
+    crypto->tkey = gcry_malloc(crypto->tkey_len);
+
+    if (!crypto->tkey)
+	return GPG_ERR_ENOMEM;
+
+    fd = open(file, O_RDONLY);
+
+    if (fd == -1)
+	return gpg_error_from_errno(errno);
+
+    gsize len = read(fd, crypto->tkey, crypto->tkey_len);
+
+    if (len != crypto->tkey_len) 
+	rc = GPG_ERR_INCOMPLETE_LINE;
+
+    close(fd);
+
+    // Converting.
+    if (crypto->fh && (crypto->fh->v1 || crypto->fh->ver.fh2.version <= 0x213)) {
+	unsigned char *p = crypto->tkey;
+	unsigned i;
+
+	for (i = 0; i < crypto->tkey_len; i++) {
+	    if (p[i] == '\0' || p[i] == '\n')
+		break;
+	}
+
+	if (i != crypto->tkey_len)
+	    log_write(N_("%s: warning: key file data truncated at byte %i."), file, i);
+
+	crypto->tkey_len = i;
     }
 
-    p = fgets(buf, sizeof(buf), fp);
-    fclose(fp);
-    len = strlen(buf);
-
-    if (len && buf[len - 1] == '\n')
-	buf[--len] = 0;
-
-    str = gcry_malloc(len + 1);
-
-    if (!str) {
-	*rc = gpg_error_from_errno(ENOMEM);
-	return FALSE;
-    }
-
-    memcpy(str, buf, len ? len : 1);
-    str[len] = 0;
-    memset(&buf, 0, sizeof(buf));
-    *result = str;
-    return TRUE;
+    return rc;
 }
 
-gchar *parse_rcfile_keyfile(const gchar *filename, gboolean import,
-	gpg_error_t *rc)
+gpg_error_t parse_rcfile_keyfile(struct crypto_s *crypto, const gchar *filename,
+	gboolean import)
 {
     GError *rv = NULL;
-    gchar *t, *file = NULL, *str;
-
-    *rc = GPG_ERR_UNKNOWN_ERRNO;
+    gchar *t, *file = NULL;
+    gpg_error_t rc = 0;
 
+    // parse_rcfile_keys()
     if (import == FALSE) {
 	if (g_key_file_has_key(keyfileh, filename, "key_file", &rv) == TRUE) {
 	    file = g_key_file_get_string(keyfileh, filename, "key_file", &rv);
@@ -101,20 +115,30 @@ gchar *parse_rcfile_keyfile(const gchar *filename, gboolean import,
 		    g_clear_error(&rv);
 		}
 
-		return NULL;
+		return GPG_ERR_ENFILE;
 	    }
 
 	    t = expand_homedir(file);
 
 	    if (!t) {
+		g_free(file);
 		log_write("%s(%i): %s", __FILE__, __LINE__, strerror(ENOMEM));
-		*rc = gpg_error_from_errno(ENOMEM);
-		return NULL;
+		return GPG_ERR_ENOMEM;
 	    }
 
 	    g_free(file);
 	    file = t;
 	}
+
+	if (rv) {
+	    log_write("%s: key_file: %s", rcfile, rv->message);
+
+	    if (file)
+		g_free(file);
+
+	    g_clear_error(&rv);
+	    return GPG_ERR_UNKNOWN_ERRNO;
+	}
     }
     else {
 	/* -I or -C. The filename is a key file. */
@@ -122,29 +146,20 @@ gchar *parse_rcfile_keyfile(const gchar *filename, gboolean import,
 
 	if (!file) {
 	    log_write("%s(%i): %s", __FILE__, __LINE__, strerror(ENOMEM));
-	    *rc = gpg_error_from_errno(ENOMEM);
-	    return NULL;
+	    return GPG_ERR_ENOMEM;
 	}
     }
 
-    if (rv) {
-	log_write("%s: key_file: %s", rcfile, rv->message);
-	g_clear_error(&rv);
-	return NULL;
-    }
-
     if (!file)
-	return NULL;
+	return GPG_ERR_ENFILE;
 
-    if (_getline(file, &str, rc) == FALSE) {
-	log_write("%s: %s: %s", filename, file, pwmd_strerror(*rc));
-	g_free(file);
-	return NULL;
-    }
+    rc = _getline(crypto, file);
+
+    if (rc)
+	log_write("%s: %s: %s", filename, file, pwmd_strerror(rc));
 
     g_free(file);
-    *rc = 0;
-    return str;
+    return rc;
 }
 
 gchar *get_key_file_string(const gchar *section, const gchar *what)
@@ -279,38 +294,57 @@ gboolean parse_rcfile_keys()
     groups = g_key_file_get_groups(keyfileh, &n);
 
     for (p = groups; *p; p++) {
-	GError *rc = NULL;
+	GError *rv = NULL;
 
-	if (g_key_file_has_key(keyfileh, *p, "key", &rc) == TRUE) {
-	    str = g_key_file_get_string(keyfileh, *p, "key", &rc);
+	if (!strcmp(*p, "global"))
+	    continue;
+
+	if (g_key_file_has_key(keyfileh, *p, "key", &rv) == TRUE) {
+	    str = g_key_file_get_string(keyfileh, *p, "key", &rv);
 
 	    if (!str) {
-		if (rc) {
-		    log_write("%s: key: %s", rcfile, rc->message);
-		    g_clear_error(&rc);
+		if (rv) {
+		    log_write("%s: key: %s", rcfile, rv->message);
+		    g_clear_error(&rv);
 		}
+
 		continue;
 	    }
 
-	    do_cache_push(*p, str);
+	    do_cache_push(*p, str, strlen(str));
 	    g_free(str);
 	    continue;
 	}
 
-	if (rc) {
-	    log_write("%s: key: %s", rcfile, rc->message);
-	    g_clear_error(&rc);
+	if (rv) {
+	    log_write("%s: key: %s", rcfile, rv->message);
+	    g_clear_error(&rv);
 	    continue;
 	}
 
-	gpg_error_t ret;
-	str = parse_rcfile_keyfile(*p, FALSE, &ret);
+	struct crypto_s *crypto = init_client_crypto();
+	gpg_error_t rc;
 
-	if (!str)
+	if (!crypto)
+	    return FALSE;
+
+	crypto->fh = read_file_header(*p, FALSE, &rc);
+
+	if (!crypto->fh) {
+	    log_write("%s: %s", *p, pwmd_strerror(rc));
+	    cleanup_crypto(&crypto);
 	    continue;
+	}
 
-	do_cache_push(*p, str);
-	gcry_free(str);
+	rc = parse_rcfile_keyfile(crypto, *p, FALSE);
+
+	if (rc) {
+	    cleanup_crypto(&crypto);
+	    continue;
+	}
+
+	do_cache_push(*p, crypto->tkey, crypto->tkey_len);
+	cleanup_crypto(&crypto);
     }
 
     g_strfreev(groups);
